/*
 * popa3d version information.
 */

char popa3d_version[] = "1.0";
char popa3d_date[] = "$Date: 2005/05/26 16:01:15 $";
char popa3d_id[] = "@(#) $Id: version.c,v 1.5 2005/05/26 16:01:15 solar Exp $";
